package com.perforce.hws.automation;

import com.perforce.hws.deployment.PathConventions;
import com.perforce.hws.util.ProcessHelpers;
import com.perforce.hws.util.TimingHelpers;
import com.perforce.hwsclient.ApiClient;
import io.jsonwebtoken.impl.crypto.MacProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Base64;

import static com.perforce.hws.automation.ProcessUtil.*;

/**
 * Methods related to launching the HWS service in the background.
 *
 * Importantly, this will ignore launching HWS if it's already running. This is
 * often the case in development environments. We let our test framework just
 * use the existing instance (usually running at port 9000).
 */
public interface ManageHwsService extends
        UsesAutomationConfig,
        ProjectLocations,
        ProcessHelpers,
        NetworkUtil,
        TimingHelpers,
        PathConventions {

    Logger logger = LoggerFactory.getLogger(ManageHwsService.class);

    default void startHWS() throws IOException {
        if (isHWSRunning()) {
            logger.info("HWS already running, not starting");
            return;
        }

        String serverJar = locateServerJar();

        if (!Files.exists(Paths.get(serverJar))) {
            throw new IOException("Did not locate HWS Jar at " + serverJar);
        }

        String log4jFile = log4jDebugXml();

        ProcessBuilder pb = new ProcessBuilder("java",
                "-javaagent:" + locateJaCoCoAgentJar(),
                "-Dlog4j.configurationFile=" + log4jFile,
                "-DHWS_CONFIG_PATH=" + config().getHwsConfig(),
                "-jar", serverJar);

        pb.directory(new File(config().getInstallDir()));

        pb.redirectInput(new File(nullPath()));

        StringBuilder configSb = new StringBuilder();

        configSb.append("P4DCONFIGDIR: ").append(config().getHwsP4dConfigDir()).append("\n");

        configSb.append("JWT_SIGNING_KEY: ")
                .append(new String(Base64.getEncoder().encode(MacProvider.generateKey().getEncoded())))
                .append("\n");

        configSb.append("PROJECTSPATH: //depot/main\n");
        configSb.append("HWS_AUTH_P4D: localhost\n");
        if (config().isCloudEnabled()) {
            configSb.append("CLOUD_ENABLED: true\n");
            configSb.append("HELIX_CLOUD_URL: http://localhost:3000\n");
        }
        if (config().isGitFusionEnabled()) {
            configSb.append("ENABLE_GIT_FUSION: true\n");
        }
        if (config().isEnableP4dTls()) {
            configSb.append("TRUST_FINGERPRINTS: " + trustFingerprintsPath() + "\n");
            Path fpPath = Paths.get(trustFingerprintsPath());
            if (!fpPath.getParent().toFile().exists()) {
                Files.createDirectories(fpPath.getParent());
            }
            Files.write(fpPath, (config().getP4dFingerprint() + "\n").getBytes());
        }

        Path configPath = Paths.get(config().getHwsConfig());

        if (!Files.exists(configPath.getParent())) {
            Files.createDirectories(configPath.getParent());
        }
        Files.write(configPath, configSb.toString().getBytes());

        Path p4dConfigPath = Paths.get(config().getHwsP4dConfigDir());
        if (!Files.exists(p4dConfigPath)) {
            Files.createDirectories(p4dConfigPath);
        }

        StringBuilder localSb = new StringBuilder();

        localSb.append("id: localhost\n");
        localSb.append("name: Local Server\n");
        localSb.append("description: A test p4d running locally\n");
        if (config().isEnableP4dTls()) {
            localSb.append("P4PORT: ssl:localhost:1666\n");
        } else {
            localSb.append("P4PORT: localhost:1666\n");
        }
        if (config().isEnableUnicodeP4d()) {
            localSb.append("P4CHARSET: utf8\n");
        }
        localSb.append("APILEVEL: 80\n");

        Path localPath = Paths.get(p4dConfigPath.toString(), "localhost");
        Files.write(localPath, localSb.toString().getBytes());

        Process launched = pb.start();

        HWSOutputLogger stderrLogger = new HWSOutputLogger(launched.getErrorStream());
        HWSOutputLogger stdoutLogger = new HWSOutputLogger(launched.getInputStream());
        stderrLogger.start();
        stdoutLogger.start();

        int pid = pidOf(launched);

        if (pid != -1) {
            logger.info("Started HWS at {}", pid);
            if (!launched.isAlive()) {
                throw new IllegalStateException("HWS start failed with exit " +
                        launched.exitValue() + ": " + pb.command());
            }
            try (PrintWriter pw = new PrintWriter(config().getHwsPidPath())) {
                pw.write(Integer.toString(pid));
            }
        } else {
            throw new IllegalStateException("Unable to launch or detect pid " +
                    "of hws process: " + pb.command());
        }

        waitForHwsToStart();
    }

    default void waitForHwsToStart() {
        try {
            int retries = 60 * 10; // Wait no more than 1 minute
            while (0 < retries && !isHWSRunning()) {
                Thread.sleep(100);
                retries--;
            }
        } catch (InterruptedException e) {
            throw new IllegalStateException(e);
        }
    }

    class HWSOutputLogger extends Thread {

        private InputStream is;

        public HWSOutputLogger(InputStream is) {
            this.is = is;
        }

        @Override
        public void run() {
            try {
                InputStreamReader isr = new InputStreamReader(is);
                BufferedReader br = new BufferedReader(isr);
                String line = null;
                while ((line = br.readLine()) != null) {
                    logger.debug(line);
                }
            } catch (IOException e) {
                logger.error("Problem reading from p4d", e);
            }
            logger.info("stopping reading of p4d");
        }
    }

    default void stopHWS() throws IOException, InterruptedException {
        String pid = getHWSPid();
        if (pid != null) {
            if (isWindows()) {
                logger.debug("Stopping hws, taskkill /F /PID {}", pid);
                String taskkill = Paths.get(System.getenv("windir"), "system32", "taskkill.exe").toString();
                Runtime.getRuntime().exec(taskkill + " /F /PID " + pid);
            } else {
                logger.debug("Stopping hws, sending SIGTERM to {}", pid);
                Runtime.getRuntime().exec("kill -SIGTERM " + pid);
            }
            waitForPidToQuit(pid);
            if (isRunning(pid) && !isWindows()) {
                logger.debug("HWS still running, sending SIGKILL to {}", pid);
                Runtime.getRuntime().exec("kill -SIGKILL " + pid);
            }
            if (isRunning(pid)) {
                throw new IllegalStateException("process " + pid + " is still running");
            }
        }
    }

    default String getHWSPid() throws IOException {
        Path pidPath = Paths.get(config().getHwsPidPath());
        if (Files.exists(pidPath)) {
            byte[] content = Files.readAllBytes(pidPath);
            return new String(content, "UTF-8");
        }
        return null;
    }

    default boolean isHWSRunning() {
        ApiClient apiClient = new ApiClient(true, config().getLocalHws());
        return apiClient.isOK();
    }

    default String trustFingerprintsPath() {
        Path configDir = Paths.get(config().getHwsConfig()).getParent();
        return Paths.get(configDir.toString(), "trust_fingerprints").toString();
    }

    /**
     * Returns the path to our server "fat jar".
     *
     * This will be what we use when we launch HWS in the background.
     *
     * @return Absolute path to the HWS jar file.
     */
    default String locateServerJar() throws IOException {
        String root = projectRootOrFail();
        // I'm not sure we want to do a glob search for this or not.
        return Paths.get(root, "server", "build", "libs", "hws.jar").toString();
    }

    default String log4jDebugXml() throws IOException {
        String root = projectRootOrFail();
        return Paths.get(root, "server", "debug-log4j.xml").toAbsolutePath().toString();
    }

    /**
     * The absolute path to the JaCoCo (code coverage) agent jar.
     *
     * @return
     */
    default String locateJaCoCoAgentJar() throws IOException {
        String root = projectRootOrFail();
        return Paths.get(root, "automation", "jacoco-0.7.7-20160421.182301-14",
                "lib", "jacocoagent.jar").toAbsolutePath().toString();
    }

    default void restartHwsService() {
        try {
            if (isWindows()) {
                String scPath = Paths.get(System.getenv("windir"), "system32", "sc.exe").toString();
                logger.info(execOutOrFail(scPath, "stop", "helix-ws"));
                waitFor(10, 1000, () -> isPortAvailable(443));
                logger.info(execOutOrFail(scPath, "start", "helix-ws"));
            } else if (isLinux()) {
                logger.info(execOutOrFail("service", "helix-ws", "stop"));
                logger.info(execOutOrFail("service", "helix-ws", "start"));
                waitForHwsToStart();
            } else if (isOSX()) {
                if (Files.exists(plistPath())) {
                    logger.info(execOutOrFail("launchctl", "unload", "-w",
                    		plistPath().toString()));
                    logger.info(execOutOrFail("launchctl", "load", "-w",
                    		plistPath().toString()));
                    waitForHwsToStart();
                } else {
                    logger.info("not restarting HWS");
                }
            } else {
                throw new IllegalStateException("Do not know how to restart HWS on this platform");
            }
        } catch (IOException io) {
            throw new UncheckedIOException(io);
        } catch (InterruptedException ie) {
            throw new RuntimeException(ie);
        }
    }
}
