package com.perforce.hws.automation;

import com.google.gson.Gson;
import com.perforce.hws.p4base.CloseableServer;
import com.perforce.hws.p4base.P4Methods;
import com.perforce.hws.p4base.ResultMap;
import com.perforce.hws.p4base.TemporaryClient;
import com.perforce.p4java.exception.P4JavaException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static java.nio.file.StandardOpenOption.APPEND;
import static java.nio.file.StandardOpenOption.CREATE;

/**
 * Created by tjuricek on 12/16/15.
 */
public interface P4dInitConventions extends
        P4Methods,
        UsesAutomationConfig,
        TemporaryClient,
        AutomationServerProvider {

    Logger logger = LoggerFactory.getLogger(P4dInitConventions.class);

    default String setSecuritySuperPassAndProtects() throws P4JavaException {
        String ticket;

        try (CloseableServer s = obtain(null, config().getSuperUser(), null)) {
            exec(s, "configure", "set", "security=3");

            changePassword(s, null, config().getSuperPassword());

            s.get().login(config().getSuperPassword());
            ticket = s.get().getAuthTicket();
            logger.info("super ticket: {}", ticket);

            // This will generate our protections table with the current user
            // as super.
            List<ResultMap> results = exec(s, "protect", "-o");
            ResultMap protect = results.get(0);
            // Seen in 16.2 (mainline), the protections table is actually completely
            // empty. So create a basic one.
            if (!protect.containsKey("Protections0")) {
                protect.put("Protections0", "super user super * //...");
                protect.put("Protections1", "write user * * //...");
            }
            results = execInput(s, protect, "protect", "-i");
            if (logger.isDebugEnabled()) {
                Gson gson = new Gson();
                logger.debug("results {}", gson.toJson(results));
            }
        }

        return ticket;
    }

    default void createCommonUsers(String superTicket) throws P4JavaException {
        try (CloseableServer s = obtain(null, config().getSuperUser(), superTicket)) {
            Map<String, Object> user1 = new HashMap<>();
            user1.put("User", "jdoe");
            user1.put("Email", "jdoe@perforce.com");
            user1.put("FullName", "John Doe");
            execInput(s, user1, "user", "-i", "-f");

            s.get().setUserName("jdoe");
            changePassword(s, null, "johndoe1A!");

            s.get().setUserName(config().getSuperUser());
            s.get().setAuthTicket(superTicket);

            Map<String, Object> user2 = new HashMap<>();
            user2.put("User", "mmustermann");
            user2.put("Email", "mmustermann@example.com");
            user2.put("FullName", "Max Mustermann");
            execInput(s, user2, "user", "-i", "-f");

            s.get().setUserName("mmustermann");
            changePassword(s, null, "mmustermann1A!");

            s.get().setUserName(config().getSuperUser());
            s.get().setAuthTicket(superTicket);

            Map<String, Object> user3 = new HashMap<>();
            user3.put("User", "j\\doe");
            user3.put("FullName", "Johnny Doe");
            user3.put("Email", "j\\doe@example.com");
            execInput(s, user3, "user", "-i", "-f");

            s.get().setUserName("j\\doe");
            changePassword(s, null, "johndoe1A!");

            s.get().setUserName(config().getSuperUser());
            s.get().setAuthTicket(superTicket);

            Map<String, Object> user4 = new HashMap<>();
            user4.put("User", "triggers");
            user4.put("FullName", "Trigger Account");
            user4.put("Email", "triggers@example.com");
            execInput(s, user4, "user", "-i", "-f");

            s.get().setUserName("triggers");
            changePassword(s, null, "triggers1A!");

            s.get().setUserName(config().getSuperUser());
            s.get().setAuthTicket(superTicket);
        }
    }

    default void submitMyProjectStreamData(CloseableServer s) throws P4JavaException, IOException {
        withStream(s, "//my_project/main", (root) -> {
            Path file1 = Paths.get(root, "README");
            String file1Content = "This is the readme for My Project\n";
            file1.getParent().toFile().mkdirs();
            Files.write(file1, file1Content.getBytes(), CREATE);

            Path file2 = Paths.get(root, "Experimental", "README");
            String file2Content = "This is an experimental project README.\n";
            file2.getParent().toFile().mkdirs();
            Files.write(file2, file2Content.getBytes(), CREATE);

            exec(s, "add", file1.toString());
            exec(s, "add", file2.toString());

            exec(s, "submit", "-d", "A few basic adds");

            exec(s, "edit", file1.toString());
            String file1Edit = "The README has been edited.\n";
            Files.write(file1, file1Edit.getBytes(), APPEND);
            exec(s, "submit", "-d", "Edits");
        });
    }

    default void createCommonDepots(String superUser, String superTicket) {
        try (CloseableServer s = obtain(null, superUser, superTicket)) {
            // This stream depot is mirrored with the Helix Cloud stream depot
            // which is used to allow us to create stream test methods on
            // both environments.

            List<ResultMap> results = exec(s, "depot", "-o", "my_project");
            ResultMap depot1 = results.get(0);
            depot1.put("Type", "stream");
            depot1.put("Description", "A cloud test depot");
            execInput(s, depot1, "depot", "-i");

            List<ResultMap> mainResults = exec(s,
                    "stream", "-o", "-t", "mainline", "//my_project/main");
            ResultMap mainStream = mainResults.get(0);
            mainStream.put("Name", "My~20Project");
            execInput(s, mainStream, "stream", "-i");

            List<ResultMap> devResults = exec(s,
                    "stream", "-o",
                    "-t", "development",
                    "-P", "//my_project/main",
                    "//my_project/dev");
            execInput(s, devResults.get(0), "stream", "-i");

        }
    }

    default void createCommonFiles(String user, String ticket) throws P4JavaException, IOException {
        try (CloseableServer s = obtain(null, user, ticket)) {
            withClient(s, (root) -> {
                Path file1 = Paths.get(root, "depot", "main", "My Project", "README");
                String file1Content = "This is the readme for My Project\n";
                file1.getParent().toFile().mkdirs();
                Files.write(file1, file1Content.getBytes(), CREATE);

                Path file2 = Paths.get(root, "depot", "dev", "Experimental", "README");
                String file2Content = "This is an experimental project README.\n";
                file2.getParent().toFile().mkdirs();
                Files.write(file2, file2Content.getBytes(), CREATE);

                exec(s, "add", file1.toString());
                exec(s, "add", file2.toString());

                exec(s, "submit", "-d", "A few basic adds");

                exec(s, "edit", file1.toString());
                String file1Edit = "The README has been edited.\n";
                Files.write(file1, file1Edit.getBytes(), APPEND);
                exec(s, "submit", "-d", "Edits");
            });

            submitMyProjectStreamData(s);
        }
    }
}
