package com.perforce.hws.automation;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

/**
 * Some helper methods to identify key paths in our project source tree.
 */
public interface ProjectLocations extends OsHelpers {

    /**
     * Tries to identify the project's root directory.
     * <p>
     * This should be callable from any particular location in the tree. We
     * start by looking at our current working directory and navigate "up".
     *
     * @return The absolute path to the project's root, or null, if we can't find it.
     */
    default String projectRoot() throws IOException {

        Path dir = Paths.get("").toAbsolutePath();

        Path root = null;

        do {
            Path matched = Files.find(dir, 1, (p, a) -> p.getFileName().toString().equals("settings.gradle"))
                    .findFirst()
                    .orElse(null);

            if (matched == null) {
                dir = dir.getParent();
            } else {
                root = dir;
            }
        } while (root == null && dir != null);

        if (root != null) {
            return root.toString();
        }

        return null;
    }

    /**
     * Try to guess the root of our client using our client conventions.
     * <p>
     * This really works only if we start from the HWS source.
     *
     * @return
     */
    default String clientRoot() throws IOException {
        Path root;
        if (branchName().equals("main")) {
            root = Paths.get(branchPath(), "..", "..");
        } else {
            root = Paths.get(branchPath(), "..", "..", "..");
        }
        return root.normalize().toAbsolutePath().toString();
    }

    default String projectRootOrFail() throws IOException {
        String root = projectRoot();
        if (root == null) {
            throw new IllegalStateException("Did not locate project root");
        }
        return root;
    }

    /**
     * Tries to locate the T4 test root based on conventions.
     *
     * @return
     */
    default String t4TestRoot() throws IOException {
        return Paths.get(projectRootOrFail(), "..", "..", "p4-test")
                .normalize()
                .toAbsolutePath()
                .toString();
    }

    default String t4TestRootOrFail() throws IOException {
        String root = t4TestRoot();
        if (root == null && !Files.exists(Paths.get(root))) {
            throw new IllegalStateException("Could not locate or find " +
                    "existing p4-test directory for this project branch.");
        }
        return root;
    }

    default String t4TestBin() throws IOException {
        // There is a t4.bat to run on windows. I'm not sure if we need that or
        // not.
        return Paths.get(t4TestRootOrFail(), "bin", "t4.pl")
                .normalize()
                .toAbsolutePath()
                .toString();
    }

    default String helixSyncBuildPath() throws IOException {
        if (OsUtil.isWindows()) {
            return Paths.get(helixSyncBuildDirPath(), "HelixSync.exe").toAbsolutePath().toString();
        } else if (OsUtil.isOSX()) {
            return Paths.get(helixSyncBuildDirPath(), "helix-sync.tgz").toAbsolutePath().toString();
        } else if (OsUtil.isLinux()) {
            return Paths.get(helixSyncBuildDirPath(), "helix-sync.tgz").toAbsolutePath().toString();
        } else {
            throw new IllegalStateException("Unknown OS type for helix sync build path");
        }
    }

    default String branchName() throws IOException {
        return Paths.get(branchPath()).getFileName().toString();
    }

    default String branchPath() throws IOException {
        return Paths.get(projectRootOrFail(), "..", "..")
                .normalize()
                .toAbsolutePath()
                .toString();
    }

    default String buildDepotPath() throws IOException {
        // TODO this might change depending upon our branch.
        return Paths.get(branchPath(), "..", "..", "builds")
                .normalize()
                .toAbsolutePath()
                .toString();
    }

    default String helixSyncBuildDirPath() throws IOException {
        return Paths.get(buildDepotPath(), branchName(), "p4-bin", helixSyncBuildDirName())
                .normalize()
                .toAbsolutePath()
                .toString();
    }

    default String helixSyncBuildDirName() throws IOException {
        // Note: I am blindly assuming we are *not* building 32-bit binaries.
        if (OsUtil.isOSX()) {
            return "bin.macosx107x86_64";
        } else if (OsUtil.isWindows()) {
            return "bin.ntx64";
        } else if (OsUtil.isLinux()) {
            return "bin.linux26x86_64";
        } else {
            throw new IllegalStateException("Can't find build dir for current OS");
        }
    }

    default String coverageOutputPath() throws IOException {
        return Paths.get(projectRootOrFail(), "build", "coverageData").toAbsolutePath().toString();
    }

    /**
     * This is a debian package location we use for a "base install" that we can test upgrades
     * against.
     *
     * @return
     */
    default String baseDebUpgradePath() {
        try {
            if (isUbuntu()) {
                String version = loadUbuntuVersion();
                if (version.startsWith("12")) {
                    return Paths.get(projectRootOrFail(), "automation", "packages", "ubuntu-12",
                            "helix-web-services_2016.1.0-1387282_amd64.deb")
                            .toAbsolutePath().toString();
                } else if (version.startsWith("14") || version.startsWith("16")) {
                    return Paths.get(projectRootOrFail(), "automation", "packages", "ubuntu-14",
                            "helix-web-services_2016.1.0-1387282_amd64.deb")
                            .toAbsolutePath().toString();
                } else {
                    throw new IllegalStateException("Ubuntu verson " + version + " not supported");
                }
            } else {
                throw new IllegalStateException("OS version is not supported");
            }
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }


    /**
     * This is a debian package location we use for a "base install" that we can test upgrades
     * against.
     *
     * @return
     */
    default String baseRpmUpgradePath() {
        try {
            if (isCentOS()) {
                return Paths.get(projectRootOrFail(), "automation", "packages", "centos-6",
                        "helix-web-services-2016.1.0-1387282.x86_64.rpm")
                        .toAbsolutePath().toString();
            } else {
                throw new IllegalStateException("OS version is not supported");
            }
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }
}
